// ============================================================================
// Copyright (c) 2018 by Terasic Technologies Inc.
// ============================================================================
//
// Permission:
//
//   Terasic grants permission to use and modify this code for use
//   in synthesis for all Terasic Development Boards and Altera Development
//   Kits made by Terasic.  Other use of this code, including the selling
//   ,duplication, or modification of any portion is strictly prohibited.
//
// Disclaimer:
//
//   This VHDL/Verilog or C/C++ source code is intended as a design reference
//   which illustrates how these types of functions can be implemented.
//   It is the user's responsibility to verify their design for
//   consistency and functionality through the use of formal
//   verification methods.  Terasic provides no warranty regarding the use
//   or functionality of this code.
//
// ============================================================================
//
//  Terasic Technologies Inc
//  9F., No.176, Sec.2, Gongdao 5th Rd, East Dist, Hsinchu City, 30070. Taiwan
//
//
//                     web: http://www.terasic.com/
//                     email: support@terasic.com
//
// ============================================================================
//Date:  Mon Jul 30 08:55:59 2018
// ============================================================================

`define ENABLE_PCIE
`define ENABLE_DDR4A
`define ENABLE_DDR4B
//`define ENABLE_CRSV

module Flik_top(

      ///////// CLOCK /////////
      input              OSC_100_CLKUSR, //CLKUSR for configuration and transceiver calibration
      input              SYS_REFCLK1_p, //default 100MHz
      input              SYS_REFCLK2_p, //default 100MHz

      ///////// KEY /////////
      input              CPU_RESET_n,

      ///////// LED /////////
      output   [ 4: 1]   LED, //LED is Low-Active

`ifdef ENABLE_DDR4A
      ///////// DDR4A /////////
      input              DDR4A_REFCLK_p,
      output   [16: 0]   DDR4A_A,
      output   [ 1: 0]   DDR4A_BA,
      output             DDR4A_CK, //DDR4A_CK_p
      output             DDR4A_CK_n,
      output             DDR4A_CKE,
      inout    [ 7: 0]   DDR4A_DQS, //DDR4A_DQS_p
      inout    [ 7: 0]   DDR4A_DQS_n,
      inout    [63: 0]   DDR4A_DQ,
      output             DDR4A_CS_n, //DDR4A_CS_L
      output             DDR4A_RESET_n, //DDR4A_RESET_L
      output             DDR4A_ODT,
      output             DDR4A_ACT_n, //DDR4A_ACT_L
      input              DDR4A_ALERT_n, //DDR4A_ALERT_L
      output             DDR4A_BG,
      inout    [ 7: 0]   DDR4A_DBI_n, //DDR4A_DBI_L
      output             DDR4A_PAR,
      input              DDR4A_RZQ, //RZQ_3C
`endif /*ENABLE_DDR4A*/

`ifdef ENABLE_DDR4B
      ///////// DDR4B /////////
//      input              DDR4B_REFCLK_p,
      output   [16: 0]   DDR4B_A,
      output   [ 1: 0]   DDR4B_BA,
      output             DDR4B_CK, //DDR4B_CK_p
      output             DDR4B_CK_n,
      output             DDR4B_CKE,
      inout    [ 7: 0]   DDR4B_DQS, //DDR4B_DQS_p
      inout    [ 7: 0]   DDR4B_DQS_n,
      inout    [63: 0]   DDR4B_DQ,
      output             DDR4B_CS_n, //DDR4B_CS_L
      output             DDR4B_RESET_n, //DDR4B_RESET_L
      output             DDR4B_ODT,
      output             DDR4B_ACT_n, //DDR4B_ACT_L
      input              DDR4B_ALERT_n, //DDR4B_ALERT_L
      output             DDR4B_BG,
      inout    [ 7: 0]   DDR4B_DBI_n, //DDR4B_DBI_L
      output             DDR4B_PAR,
      input              DDR4B_RZQ, //RZQ_3F
`endif /*ENABLE_DDR4B*/

      ///////// SI5340B /////////
      inout              SI5340B_I2C_SCL,
      inout              SI5340B_I2C_SDA,
      input              SI5340B_INTR,
      output             SI5340B_OE_n,
      output             SI5340B_RST_n,

`ifdef ENABLE_PCIE
      ///////// PCIE /////////
      input              PCIE_REFCLK_p,
      output   [ 3: 0]   PCIE_TX_p,
      input    [ 3: 0]   PCIE_RX_p,
      input              PCIE_PERST_n,
      output             PCIE_WAKE_n,
`endif /*ENABLE_PCIE*/

`ifdef ENABLE_CRSV
      ///////// Clock RSV for XCVR /////////
      input              RSV_REFCLK_p,
`endif /*ENABLE_CRSV*/

      ///////// DEBUG /////////
      inout    [15: 1]   DEBUG,

      ///////// M10A10 /////////
      inout    [ 7: 0]   M10A10_R_IO

);


//=======================================================
//  REG/WIRE declarations
//=======================================================

wire ddr4_a_status_local_cal_success;
wire ddr4_a_status_local_cal_fail;

wire ddr4_b_status_local_cal_success;
wire ddr4_b_status_local_cal_fail;

//=======================================================
//  Structural coding
//=======================================================
assign SI5340B_OE_n = 1'b0;
assign SI5340B_RST_n = 1'b1;

// What assignments do I need for a PCIe Gen3 design that targets an Arria 10 ES2 or Production devices?
// https://www.altera.com/support/support-resources/knowledge-base/solutions/rd04242015_385.html

wire [31:0] pcie_a10_hip_0_hip_ctrl_test_in;

assign pcie_a10_hip_0_hip_ctrl_test_in = 32'h000000A8;
assign PCIE_WAKE_n = 1'b1;

//////////////////////
// PCIE RESET
wire             any_rstn;
reg              any_rstn_r /* synthesis ALTERA_ATTRIBUTE = "SUPPRESS_DA_RULE_INTERNAL=R102"  */;
reg              any_rstn_rr /* synthesis ALTERA_ATTRIBUTE = "SUPPRESS_DA_RULE_INTERNAL=R102"  */;

assign any_rstn = PCIE_PERST_n & CPU_RESET_n;


//reset Synchronizer
always @(posedge SYS_REFCLK1_p or negedge any_rstn)
begin
    if (any_rstn == 0)
    begin
        any_rstn_r <= 0;
        any_rstn_rr <= 0;
    end
    else
    begin
        any_rstn_r <= 1;
        any_rstn_rr <= any_rstn_r;
    end
end


ep_g3x8_avmm256_integrated u0 (
    .pcie_a10_hip_0_hip_ctrl_test_in          (pcie_a10_hip_0_hip_ctrl_test_in),    //        pcie_a10_hip_0_hip_ctrl.test_in
    .pcie_a10_hip_0_hip_ctrl_simu_mode_pipe   (1'b0),                               //                               .simu_mode_pipe
    .pcie_a10_hip_0_hip_pipe_sim_pipe_pclk_in (1'b0),                               //        pcie_a10_hip_0_hip_pipe.sim_pipe_pclk_in
    .pcie_a10_hip_0_hip_serial_rx_in0         (PCIE_RX_p[0]),                       //      pcie_a10_hip_0_hip_serial.rx_in0
    .pcie_a10_hip_0_hip_serial_rx_in1         (PCIE_RX_p[1]),                       //                               .rx_in1
    .pcie_a10_hip_0_hip_serial_rx_in2         (PCIE_RX_p[2]),                       //                               .rx_in2
    .pcie_a10_hip_0_hip_serial_rx_in3         (PCIE_RX_p[3]),                       //                               .rx_in3
    .pcie_a10_hip_0_hip_serial_tx_out0        (PCIE_TX_p[0]),                       //                               .tx_out0
    .pcie_a10_hip_0_hip_serial_tx_out1        (PCIE_TX_p[1]),                       //                               .tx_out1
    .pcie_a10_hip_0_hip_serial_tx_out2        (PCIE_TX_p[2]),                       //                               .tx_out2
    .pcie_a10_hip_0_hip_serial_tx_out3        (PCIE_TX_p[3]),                       //                               .tx_out3
    .pcie_a10_hip_0_npor_npor                 (any_rstn_rr),                        //            pcie_a10_hip_0_npor.npor
    .pcie_a10_hip_0_npor_pin_perst            (PCIE_PERST_n),                       //                               .pin_perst
    .reconfig_xcvr_clk_clk                    (SYS_REFCLK1_p),                      //              reconfig_xcvr_clk.clk
    .reconfig_xcvr_reset_reset_n              ((CPU_RESET_n==1'b0)?1'b0:(PCIE_PERST_n==1'b0)?1'b0:1'b1),
                                                                                    //            reconfig_xcvr_reset.reset_n
    .refclk_clk                               (PCIE_REFCLK_p),                      //                         refclk.clk

    .pio_led_external_connection_export       (pio_led),                            //    pio_led_external_connection.export

    .emif_ddr4_a_pll_ref_clk_clk              (DDR4A_REFCLK_p),                     //           emif_ddr4_a_pll_ref_clk.clk
    .emif_ddr4_a_oct_oct_rzqin                (DDR4A_RZQ),                          //                   emif_ddr4_a_oct.oct_rzqin
    .emif_ddr4_a_mem_mem_ck                   (DDR4A_CK),                           //                   emif_ddr4_a_mem.mem_ck
    .emif_ddr4_a_mem_mem_ck_n                 (DDR4A_CK_n),                         //                                  .mem_ck_n
    .emif_ddr4_a_mem_mem_a                    (DDR4A_A),                            //                                  .mem_a
    .emif_ddr4_a_mem_mem_act_n                (DDR4A_ACT_n),                        //                                  .mem_act_n
    .emif_ddr4_a_mem_mem_ba                   (DDR4A_BA),                           //                                  .mem_ba
    .emif_ddr4_a_mem_mem_bg                   (DDR4A_BG),                           //                                  .mem_bg
    .emif_ddr4_a_mem_mem_cke                  (DDR4A_CKE),                          //                                  .mem_cke
    .emif_ddr4_a_mem_mem_cs_n                 (DDR4A_CS_n),                         //                                  .mem_cs_n
    .emif_ddr4_a_mem_mem_odt                  (DDR4A_ODT),                          //                                  .mem_odt
    .emif_ddr4_a_mem_mem_reset_n              (DDR4A_RESET_n),                      //                                  .mem_reset_n
    .emif_ddr4_a_mem_mem_par                  (DDR4A_PAR),                          //                                  .mem_par
    .emif_ddr4_a_mem_mem_alert_n              (DDR4A_ALERT_n),                      //                                  .mem_alert_n
    .emif_ddr4_a_mem_mem_dqs                  (DDR4A_DQS),                          //                                  .mem_dqs
    .emif_ddr4_a_mem_mem_dqs_n                (DDR4A_DQS_n),                        //                                  .mem_dqs_n
    .emif_ddr4_a_mem_mem_dq                   (DDR4A_DQ),                           //                                  .mem_dq
    .emif_ddr4_a_mem_mem_dbi_n                (DDR4A_DBI_n),                        //                                  .mem_dbi_n
    .emif_ddr4_a_status_local_cal_success     (ddr4_a_status_local_cal_success),    //                emif_ddr4_a_status.local_cal_success
    .emif_ddr4_a_status_local_cal_fail        (ddr4_a_status_local_cal_fail),       //                                  .local_cal_fail
    .ddr4_a_status_external_connection_export ({ddr4_a_status_local_cal_fail,ddr4_a_status_local_cal_success}),
                                                                                    // ddr4_a_status_external_connection.export
    .emif_ddr4_b_pll_ref_clk_clk              (DDR4A_REFCLK_p),                     //           emif_ddr4_b_pll_ref_clk.clk
    .emif_ddr4_b_oct_oct_rzqin                (DDR4B_RZQ),                         //                   emif_ddr4_b_oct.oct_rzqin
    .emif_ddr4_b_mem_mem_ck                   (DDR4B_CK),                           //                   emif_ddr4_b_mem.mem_ck
    .emif_ddr4_b_mem_mem_ck_n                 (DDR4B_CK_n),                         //                                  .mem_ck_n
    .emif_ddr4_b_mem_mem_a                    (DDR4B_A),                            //                                  .mem_a
    .emif_ddr4_b_mem_mem_act_n                (DDR4B_ACT_n),                        //                                  .mem_act_n
    .emif_ddr4_b_mem_mem_ba                   (DDR4B_BA),                           //                                  .mem_ba
    .emif_ddr4_b_mem_mem_bg                   (DDR4B_BG),                           //                                  .mem_bg
    .emif_ddr4_b_mem_mem_cke                  (DDR4B_CKE),                          //                                  .mem_cke
    .emif_ddr4_b_mem_mem_cs_n                 (DDR4B_CS_n),                         //                                  .mem_cs_n
    .emif_ddr4_b_mem_mem_odt                  (DDR4B_ODT),                          //                                  .mem_odt
    .emif_ddr4_b_mem_mem_reset_n              (DDR4B_RESET_n),                      //                                  .mem_reset_n
    .emif_ddr4_b_mem_mem_par                  (DDR4B_PAR),                          //                                  .mem_par
    .emif_ddr4_b_mem_mem_alert_n              (DDR4B_ALERT_n),                      //                                  .mem_alert_n
    .emif_ddr4_b_mem_mem_dqs                  (DDR4B_DQS),                          //                                  .mem_dqs
    .emif_ddr4_b_mem_mem_dqs_n                (DDR4B_DQS_n),                        //                                  .mem_dqs_n
    .emif_ddr4_b_mem_mem_dq                   (DDR4B_DQ),                           //                                  .mem_dq
    .emif_ddr4_b_mem_mem_dbi_n                (DDR4B_DBI_n),                        //                                  .mem_dbi_n
    .emif_ddr4_b_status_local_cal_success     (ddr4_b_status_local_cal_success),    //                emif_ddr4_b_status.local_cal_success
    .emif_ddr4_b_status_local_cal_fail        (ddr4_b_status_local_cal_fail),       //                                  .local_cal_fail
    .ddr4_b_status_external_connection_export ({ddr4_b_status_local_cal_fail,ddr4_b_status_local_cal_success})
                                                                                    // ddr4_b_status_external_connection.export

);

wire [3: 0] pio_led;

reg [31: 0] por_count = 32'b0;
reg [31: 0] cpu_reset_count = 32'b0;
reg [31: 0] pcie_perst_count = 32'b0;

always @(posedge SYS_REFCLK1_p) begin
    if (por_count[25] == 1'b0) begin
        por_count = por_count + 32'b1;
    end
end

always @(posedge SYS_REFCLK1_p) begin
    if (CPU_RESET_n == 1'b0) begin
        cpu_reset_count = 32'b0;
    end
    else if (cpu_reset_count[25] == 1'b0) begin
        cpu_reset_count = cpu_reset_count + 32'b1;
    end
end

always @(posedge SYS_REFCLK1_p) begin
    if (PCIE_PERST_n == 1'b0) begin
        pcie_perst_count = 32'b0;
    end
    if (pcie_perst_count[25] == 1'b0) begin
        pcie_perst_count = pcie_perst_count + 32'b1;
    end
end

wire [3: 0] led_o;
assign led_o[0] = (~por_count[25]) ? 1'b1 : pio_led[0];
assign led_o[1] = (~cpu_reset_count[25]) ? 1'b1 : pio_led[1];
assign led_o[2] = (~pcie_perst_count[25]) ? 1'b1 : pio_led[2];
assign led_o[3] = (~por_count[25]) ? 1'b1 : pio_led[3];

assign LED = ~led_o; // led low-active

endmodule
